@@ -2,7 +2,6 @@
 
 """Test case related stuff."""
 
-__metaclass__ = type
 __all__ = [
     'attr',
     'clone_test_with_new_id',
@@ -13,6 +12,7 @@ __all__ = [
     'skipIf',
     'skipUnless',
     'TestCase',
+    'unique_text_generator',
     ]
 
 import copy
@@ -21,23 +21,14 @@ import itertools
 import sys
 import types
 import unittest
+import warnings
 
-from extras import (
-    safe_hasattr,
-    try_import,
-    )
-
-from testtools import (
-    content,
-    )
-from testtools.compat import (
-    advance_iterator,
-    reraise,
-    )
+from testtools.compat import reraise
+from testtools import content
+from testtools.helpers import try_import
 from testtools.matchers import (
     Annotate,
     Contains,
-    Equals,
     MatchesAll,
     MatchesException,
     MismatchError,
@@ -46,18 +37,20 @@ from testtools.matchers import (
     Not,
     Raises,
     )
+from testtools.matchers._basic import _FlippedEquals
 from testtools.monkey import patch
-from testtools.runtest import RunTest
+from testtools.runtest import (
+    MultipleExceptions,
+    RunTest,
+)
 from testtools.testresult import (
     ExtendedToOriginalDecorator,
     TestResult,
-    )
+)
 
-wraps = try_import('functools.wraps')
 
 class TestSkipped(Exception):
     """Raised within TestCase.run() when a test is skipped."""
-TestSkipped = try_import('unittest2.case.SkipTest', TestSkipped)
 TestSkipped = try_import('unittest.case.SkipTest', TestSkipped)
 
 
@@ -67,19 +60,16 @@ class _UnexpectedSuccess(Exception):
     Note that this exception is private plumbing in testtools' testcase
     module.
     """
-_UnexpectedSuccess = try_import(
-    'unittest2.case._UnexpectedSuccess', _UnexpectedSuccess)
 _UnexpectedSuccess = try_import(
     'unittest.case._UnexpectedSuccess', _UnexpectedSuccess)
 
+
 class _ExpectedFailure(Exception):
-    """An expected failure occured.
+    """An expected failure occurred.
 
     Note that this exception is private plumbing in testtools' testcase
     module.
     """
-_ExpectedFailure = try_import(
-    'unittest2.case._ExpectedFailure', _ExpectedFailure)
 _ExpectedFailure = try_import(
     'unittest.case._ExpectedFailure', _ExpectedFailure)
 
@@ -159,6 +149,9 @@ def _copy_content(content_object):
 def gather_details(source_dict, target_dict):
     """Merge the details from ``source_dict`` into ``target_dict``.
 
+    ``gather_details`` evaluates all details in ``source_dict``. Do not use it
+    if the details are not ready to be evaluated.
+
     :param source_dict: A dictionary of details will be gathered.
     :param target_dict: A dictionary into which details will be gathered.
     """
@@ -166,11 +159,55 @@ def gather_details(source_dict, target_dict):
         new_name = name
         disambiguator = itertools.count(1)
         while new_name in target_dict:
-            new_name = '%s-%d' % (name, advance_iterator(disambiguator))
+            new_name = '%s-%d' % (name, next(disambiguator))
         name = new_name
         target_dict[name] = _copy_content(content_object)
 
 
+# Circular import: fixtures imports gather_details from here, we import
+# fixtures, leading to gather_details not being available and fixtures being
+# unable to import it.
+fixtures = try_import('fixtures')
+
+
+def _mods(i, mod):
+    (q, r) = divmod(i, mod)
+    while True:
+        yield r
+        if not q:
+            break
+        (q, r) = divmod(q, mod)
+
+
+def _unique_text(base_cp, cp_range, index):
+    s = ''
+    for m in _mods(index, cp_range):
+        s += chr(base_cp + m)
+    return s
+
+
+def unique_text_generator(prefix):
+    """Generates unique text values.
+
+    Generates text values that are unique. Use this when you need arbitrary
+    text in your test, or as a helper for custom anonymous factory methods.
+
+    :param prefix: The prefix for text.
+    :return: text that looks like '<prefix>-<text_with_unicode>'.
+    :rtype: str
+    """
+    # 0x1e00 is the start of a range of glyphs that are easy to see are
+    # unicode since they've got circles and dots and other diacriticals.
+    # 0x1eff is the end of the range of these diacritical chars.
+    BASE_CP = 0x1e00
+    CP_RANGE = 0x1f00 - BASE_CP
+    index = 0
+    while True:
+        unique_text = _unique_text(BASE_CP, CP_RANGE, index)
+        yield f'{prefix}-{unique_text}'
+        index = index + 1
+
+
 class TestCase(unittest.TestCase):
     """Extensions to the basic TestCase.
 
@@ -198,17 +235,8 @@ class TestCase(unittest.TestCase):
             ``TestCase.run_tests_with`` if given.
         """
         runTest = kwargs.pop('runTest', None)
-        super(TestCase, self).__init__(*args, **kwargs)
-        self._cleanups = []
-        self._unique_id_gen = itertools.count(1)
-        # Generators to ensure unique traceback ids.  Maps traceback label to
-        # iterators.
-        self._traceback_id_gens = {}
-        self.__setup_called = False
-        self.__teardown_called = False
-        # __details is lazy-initialized so that a constructed-but-not-run
-        # TestCase is safe to use with clone_test_with_new_id.
-        self.__details = None
+        super().__init__(*args, **kwargs)
+        self._reset()
         test_method = self._get_test_method()
         if runTest is None:
             runTest = getattr(
@@ -228,15 +256,30 @@ class TestCase(unittest.TestCase):
             (Exception, self._report_error),
             ]
 
+    def _reset(self):
+        """Reset the test case as if it had never been run."""
+        self._cleanups = []
+        self._unique_id_gen = itertools.count(1)
+        # Generators to ensure unique traceback ids.  Maps traceback label to
+        # iterators.
+        self._traceback_id_gens = {}
+        self.__setup_called = False
+        self.__teardown_called = False
+        # __details is lazy-initialized so that a constructed-but-not-run
+        # TestCase is safe to use with clone_test_with_new_id.
+        self.__details = None
+
     def __eq__(self, other):
         eq = getattr(unittest.TestCase, '__eq__', None)
         if eq is not None and not unittest.TestCase.__eq__(self, other):
             return False
-        return self.__dict__ == other.__dict__
+        return self.__dict__ == getattr(other, '__dict__', None)
+
+    __hash__ = unittest.TestCase.__hash__
 
     def __repr__(self):
         # We add id to the repr because it makes testing testtools easier.
-        return "<%s id=0x%0x>" % (self.id(), id(self))
+        return f"<{self.id()} id=0x{id(self):0x}>"
 
     def addDetail(self, name, content_object):
         """Add a detail to be reported with this test's outcome.
@@ -289,9 +332,12 @@ class TestCase(unittest.TestCase):
         """
         raise self.skipException(reason)
 
-    # skipTest is how python2.7 spells this. Sometime in the future
-    # This should be given a deprecation decorator - RBC 20100611.
-    skip = skipTest
+    def skip(self, reason):
+        """DEPRECATED: Use skipTest instead."""
+        warnings.warn(
+            'Only valid in 1.8.1 and earlier. Use skipTest instead.',
+            DeprecationWarning, stacklevel=2)
+        self.skipTest(reason)
 
     def _formatTypes(self, classOrIterable):
         """Format a class or a bunch of classes for display in an error."""
@@ -328,7 +374,7 @@ class TestCase(unittest.TestCase):
         the exception has caused.
 
         Handlers are called in first-added, first-called order, and if they
-        raise an exception, that will propogate out of the test running
+        raise an exception, that will propagate out of the test running
         machinery, halting test processing. As a result, do not call code that
         may unreasonably fail.
         """
@@ -344,7 +390,7 @@ class TestCase(unittest.TestCase):
         :param observed: The observed value.
         :param message: An optional message to include in the error.
         """
-        matcher = Equals(expected)
+        matcher = _FlippedEquals(expected)
         self.assertThat(observed, matcher, message)
 
     failUnlessEqual = assertEquals = assertEqual
@@ -406,16 +452,18 @@ class TestCase(unittest.TestCase):
            deemed to have suffered an error, exactly as for an
            unexpected exception.
         """
-        class ReRaiseOtherTypes(object):
+        class ReRaiseOtherTypes:
             def match(self, matchee):
                 if not issubclass(matchee[0], excClass):
                     reraise(*matchee)
-        class CaptureMatchee(object):
+
+        class CaptureMatchee:
             def match(self, matchee):
                 self.matchee = matchee[1]
         capture = CaptureMatchee()
-        matcher = Raises(MatchesAll(ReRaiseOtherTypes(),
-                MatchesException(excClass), capture))
+        matcher = Raises(
+            MatchesAll(ReRaiseOtherTypes(),
+                       MatchesException(excClass), capture))
         our_callable = Nullary(callableObj, *args, **kwargs)
         self.assertThat(our_callable, matcher)
         return capture.matchee
@@ -432,6 +480,7 @@ class TestCase(unittest.TestCase):
         if mismatch_error is not None:
             raise mismatch_error
 
+    assertItemsEqual = unittest.TestCase.assertCountEqual
     def addDetailUniqueName(self, name, content_object):
         """Add a detail to the test, but ensure it's name is unique.
 
@@ -457,13 +506,14 @@ class TestCase(unittest.TestCase):
         """Check that matchee is matched by matcher, but delay the assertion failure.
 
         This method behaves similarly to ``assertThat``, except that a failed
-        match does not exit the test immediately. The rest of the test code will
-        continue to run, and the test will be marked as failing after the test
-        has finished.
+        match does not exit the test immediately. The rest of the test code
+        will continue to run, and the test will be marked as failing after the
+        test has finished.
 
         :param matchee: An object to match with matcher.
         :param matcher: An object meeting the testtools.Matcher protocol.
         :param message: If specified, show this message with any failed match.
+
         """
         mismatch_error = self._matchHelper(matchee, matcher, message, verbose)
 
@@ -527,7 +577,7 @@ class TestCase(unittest.TestCase):
         Use this when you need an arbitrary integer in your test, or as a
         helper for custom anonymous factory methods.
         """
-        return advance_iterator(self._unique_id_gen)
+        return next(self._unique_id_gen)
 
     def getUniqueString(self, prefix=None):
         """Get a string unique to this test.
@@ -545,12 +595,12 @@ class TestCase(unittest.TestCase):
         return '%s-%d' % (prefix, self.getUniqueInteger())
 
     def onException(self, exc_info, tb_label='traceback'):
-        """Called when an exception propogates from test code.
+        """Called when an exception propagates from test code.
 
         :seealso addOnException:
         """
         if exc_info[0] not in [
-            TestSkipped, _UnexpectedSuccess, _ExpectedFailure]:
+                self.skipException, _UnexpectedSuccess, _ExpectedFailure]:
             self._report_traceback(exc_info, tb_label=tb_label)
         for handler in self.__exception_handlers:
             handler(exc_info)
@@ -580,18 +630,21 @@ class TestCase(unittest.TestCase):
         id_gen = self._traceback_id_gens.setdefault(
             tb_label, itertools.count(0))
         while True:
-            tb_id = advance_iterator(id_gen)
+            tb_id = next(id_gen)
             if tb_id:
                 tb_label = '%s-%d' % (tb_label, tb_id)
             if tb_label not in self.getDetails():
                 break
-        self.addDetail(tb_label, content.TracebackContent(exc_info, self))
+        self.addDetail(tb_label, content.TracebackContent(
+            exc_info, self, capture_locals=getattr(
+                self, '__testtools_tb_locals__', False)))
 
     @staticmethod
     def _report_unexpected_success(self, result, err):
         result.addUnexpectedSuccess(self, details=self.getDetails())
 
     def run(self, result=None):
+        self._reset()
         try:
             run_test = self.__RunTest(
                 self, self.exception_handlers, last_resort=self._report_error)
@@ -660,10 +713,24 @@ class TestCase(unittest.TestCase):
         """
         try:
             fixture.setUp()
+        except MultipleExceptions as e:
+            if (fixtures is not None and
+                    e.args[-1][0] is fixtures.fixture.SetupError):
+                gather_details(e.args[-1][1].args[0], self.getDetails())
+            raise
         except:
             exc_info = sys.exc_info()
             try:
-                gather_details(fixture.getDetails(), self.getDetails())
+                # fixture._details is not available if using the newer
+                # _setUp() API in Fixtures because it already cleaned up
+                # the fixture.  Ideally this whole try/except is not
+                # really needed any more, however, we keep this code to
+                # remain compatible with the older setUp().
+                if (
+                    hasattr(fixture, '_details') and
+                    fixture._details is not None
+                ):
+                    gather_details(fixture.getDetails(), self.getDetails())
             except:
                 # Report the setUp exception, then raise the error during
                 # gather_details.
@@ -680,7 +747,7 @@ class TestCase(unittest.TestCase):
             return fixture
 
     def setUp(self):
-        super(TestCase, self).setUp()
+        super().setUp()
         if self.__setup_called:
             raise ValueError(
                 "In File: %s\n"
@@ -691,7 +758,7 @@ class TestCase(unittest.TestCase):
         self.__setup_called = True
 
     def tearDown(self):
-        super(TestCase, self).tearDown()
+        super().tearDown()
         if self.__teardown_called:
             raise ValueError(
                 "In File: %s\n"
@@ -702,7 +769,7 @@ class TestCase(unittest.TestCase):
         self.__teardown_called = True
 
 
-class PlaceHolder(object):
+class PlaceHolder:
     """A placeholder test.
 
     `PlaceHolder` implements much of the same interface as TestCase and is
@@ -741,7 +808,7 @@ class PlaceHolder(object):
         internal = [self._outcome, self._test_id, self._details]
         if self._short_description is not None:
             internal.append(self._short_description)
-        return "<%s.%s(%s)>" % (
+        return "<{}.{}({})>".format(
             self.__class__.__module__,
             self.__class__.__name__,
             ", ".join(map(repr, internal)))
@@ -794,7 +861,8 @@ def ErrorHolder(test_id, error, short_description=None, details=None):
     :param short_description: An optional short description of the test.
     :param details: Outcome details as accepted by addSuccess etc.
     """
-    return PlaceHolder(test_id, short_description=short_description,
+    return PlaceHolder(
+        test_id, short_description=short_description,
         details=details, outcome='addError', error=error)
 
 
@@ -830,14 +898,14 @@ def attr(*args):
         alter its id to enumerate the added attributes.
     """
     def decorate(fn):
-        if not safe_hasattr(fn, '__testtools_attrs'):
+        if not hasattr(fn, '__testtools_attrs'):
             fn.__testtools_attrs = set()
         fn.__testtools_attrs.update(args)
         return fn
     return decorate
 
 
-class WithAttributes(object):
+class WithAttributes:
     """A mix-in class for modifying test id by attributes.
 
     e.g.
@@ -850,7 +918,7 @@ class WithAttributes(object):
     """
 
     def id(self):
-        orig = super(WithAttributes, self).id()
+        orig = super().id()
         # Depends on testtools.TestCase._get_test_method, be nice to support
         # plain unittest.
         fn = self._get_test_method()
@@ -860,6 +928,12 @@ class WithAttributes(object):
         return orig + '[' + ','.join(sorted(attributes)) + ']'
 
 
+class_types = [type]
+if getattr(types, 'ClassType', None) is not None:
+    class_types.append(types.ClassType)
+class_types = tuple(class_types)
+
+
 def skip(reason):
     """A decorator to skip unit tests.
 
@@ -868,20 +942,19 @@ def skip(reason):
     @unittest.skip decorator.
     """
     def decorator(test_item):
+        if not isinstance(test_item, class_types):
+            @functools.wraps(test_item)
+            def skip_wrapper(*args, **kwargs):
+                raise TestCase.skipException(reason)
+            test_item = skip_wrapper
+
         # This attribute signals to RunTest._run_core that the entire test
         # must be skipped - including setUp and tearDown. This makes us
         # compatible with testtools.skip* functions, which set the same
         # attributes.
         test_item.__unittest_skip__ = True
         test_item.__unittest_skip_why__ = reason
-        if wraps is not None:
-            @wraps(test_item)
-            def skip_wrapper(*args, **kwargs):
-                raise TestCase.skipException(reason)
-        else:
-            def skip_wrapper(test_item):
-                test_item.skip(reason)
-        return skip_wrapper
+        return test_item
     return decorator
 
 
@@ -889,6 +962,7 @@ def skipIf(condition, reason):
     """A decorator to skip a test if the condition is true."""
     if condition:
         return skip(reason)
+
     def _id(obj):
         return obj
     return _id
@@ -898,6 +972,7 @@ def skipUnless(condition, reason):
     """A decorator to skip a test unless the condition is true."""
     if not condition:
         return skip(reason)
+
     def _id(obj):
         return obj
     return _id
@@ -949,7 +1024,7 @@ class ExpectedException:
         return True
 
 
-class Nullary(object):
+class Nullary:
     """Turn a callable into a nullary callable.
 
     The advantage of this over ``lambda: f(*args, **kwargs)`` is that it
@@ -968,7 +1043,7 @@ class Nullary(object):
         return repr(self._callable_object)
 
 
-class DecorateTestCaseResult(object):
+class DecorateTestCaseResult:
     """Decorate a TestCase and permit customisation of the result for runs."""
 
     def __init__(self, case, callout, before_run=None, after_run=None):
